//----------------------------------------------------------------------
//MODEULE: at45db081.c
//COMMON DESCRIPTION:
//  This module provide FLASH ROM support for chip AT45DB081
//  via SPI interface.
//  Module SPI must be enabled.
//----------------------------------------------------------------------
#include "at45db081.h"
//----------------------------------------------------------------------
#ifdef AT45DB081_ENABLE
//----------------------------------------------------------------------
// File system header (at adress 8 of FLASH ROM contains)
//----------------------------------------------------------------------
typedef struct
    {char		marker[0x10]; // Marker (<= 16chars).
     char		fsver[0x10];  // Version (<= 16chars).
     signed int		last_wpage;   // Last writting page (for cyclic regeneration)
     signed int 	curpage;      // Current page (page in cache)
     signed long	fptr;	      // File pointer
     signed long	fsize;	      // File size (up to 2GB :))
     signed int		flags;	      // flags
     //
    }sAT45DB081_MBR;
//----------------------------------------------------------------------
// Definition flags (sAT45DB081_MBR.flags)
//----------------------------------------------------------------------
// =1, - flash opened
#define	fAT45DB081_OPENED	0x0001
// =1, - needs write cache to flash
#define	fAT45DB081_NEEDSYNC	0x0002
//----------------------------------------------------------------------
// System info for flash.
// 
sAT45DB081_MBR at45db081_mbr;
//----------------------------------------------------------------------
// Cache structure.
// Page - 0x08  bytes - system info,
// 	  0x100 bytes - data
//----------------------------------------------------------------------
typedef struct
    {// 8 bytes - header of page
     unsigned int  pagenum;  // 2 bytes - number of page
     unsigned int  crc_data; // 2 bytes - crc16 of dpage
     unsigned char res0[0x0004]; // 4 reserved bytes
     //
     unsigned char dpage[0x0100]; // Data of page
    }sAT45DB081_PAGE;
//
sAT45DB081_PAGE at45db081_pcache;
//----------------------------------------------------------------------
// Wait for chip ready
//----------------------------------------------------------------------
void at45db081_wait()
  {unsigned char snd=0xD7;
   unsigned char rcv=0;
   // Wait for bit 7 (RDY_BUSY) = 1
   while(!(rcv & 0x80)){AT45DB081_SPI_TRANSFER(&snd, 1, 0, 0, &rcv, 1);}
  }
//----------------------------------------------------------------------
// Returns adress of page in transfer format
//----------------------------------------------------------------------
unsigned long at45db081_fcmd(unsigned char opcode, unsigned int n)
  {unsigned long l;
   ((char*)(&l))[0] = opcode; // must be opcode 
   ((char*)(&l))[1] = (n >> 7) & 0x1F; // PA11-PA7
   ((char*)(&l))[2] = (n << 1) & 0xFE; // PA6-PA0
   ((char*)(&l))[3] = 0;
   //
   return(l);
  }
//----------------------------------------------------------------------
// Read page n to cache trought buffer1
// Modify: at45db081_mbr.curpage
//----------------------------------------------------------------------
void at45db081_rd_page_buf1(unsigned int n)
    {// Read data from flash to buffer
     unsigned long l=at45db081_fcmd(0x53,n);
     AT45DB081_SPI_TRANSFER((char*)&l, sizeof(l), 0, 0, 0, 0);
     at45db081_wait();
     // Read data from buffer to cache
     l=at45db081_fcmd(0x54,0);
     AT45DB081_SPI_TRANSFER((char*)&l, sizeof(l), (char*)&l, 1, (char*)&at45db081_pcache, sizeof(at45db081_pcache));
     at45db081_wait();
     // Current page
     at45db081_mbr.curpage=n;
    }
//----------------------------------------------------------------------
// Write page n from cache to chip trought buffer1
//----------------------------------------------------------------------
void at45db081_wr_page_buf1(unsigned int n)
    {// Write data from cache to buffer
     unsigned long l=at45db081_fcmd(0x84,0);
     AT45DB081_SPI_TRANSFER((char*)&l, sizeof(l), (char*)&at45db081_pcache, sizeof(at45db081_pcache),0,0);
     at45db081_wait();
     // Write buffer to flash
     l=at45db081_fcmd(0x83,n);
     AT45DB081_SPI_TRANSFER((char*)&l, sizeof(l), 0, 0, 0, 0);
     at45db081_wait();
     //------------------------------------------------------------------
     // Refresh next page !!!
     //------------------------------------------------------------------
     // Calculate number of next refresh page
     if(at45db081_mbr.last_wpage==n){at45db081_mbr.last_wpage++;}
     if(at45db081_mbr.last_wpage>=AT45DB081_PAGES){at45db081_mbr.last_wpage=0;}
     // Read  page to buffer2
     l=at45db081_fcmd(0x55,at45db081_mbr.last_wpage);
     AT45DB081_SPI_TRANSFER((char*)&l, sizeof(l), 0, 0, 0, 0);
     at45db081_wait();
     // Write page from buffer2 to flash
     l=at45db081_fcmd(0x86,at45db081_mbr.last_wpage);
     AT45DB081_SPI_TRANSFER((char*)&l, sizeof(l), 0, 0, 0, 0);
     at45db081_wait();
     // go to next page
     at45db081_mbr.last_wpage++;
     //------------------------------------------------------------------
    }    
//----------------------------------------------------------------------
// Format flash
// Create structures of filesystem description
// and write it to page 0
// After this operation flash file has zero len.
//----------------------------------------------------------------------
void at45db081_format_mbr()
  {int i;
   // Marker (<= 16chars).
   strcpy_P(at45db081_mbr.marker,PSTR(FS_MARKER));
   // Version (<= 16chars). "v.00.00.00"
   strcpy_P(at45db081_mbr.fsver,PSTR(FS_VERSION));
   // Last writting page (for cyclic regeneration)
   at45db081_mbr.last_wpage = 0x00;
   // Current page *
   at45db081_mbr.curpage=0x00;
   // File pointer *
   at45db081_mbr.fptr=0x00;
   // File size *
   at45db081_mbr.fsize=0x00;
   // Flags *
   at45db081_mbr.flags=0x00;
   // Copy to cache
   for(i=0;i<sizeof(at45db081_mbr);i++){at45db081_pcache.dpage[i]=((char*)(&at45db081_mbr))[i];}
   // Page 0 sysinfo
   at45db081_pcache.pagenum=0;
   at45db081_pcache.crc_data=0;
   at45db081_pcache.res0[0]='r';
   at45db081_pcache.res0[1]='e';
   at45db081_pcache.res0[2]='s';
   at45db081_pcache.res0[3]='v';
   // Write to page 0x0000
   at45db081_wr_page_buf1(0);
  }
//----------------------------------------------------------------------
// Read mbr and check flash
// If flash is not formatted or flash already opened, returns -1,
// else (if all ok) - returns 0, file pointer set on first byte of 
// flash file.
//----------------------------------------------------------------------
signed int at45db081_open()
  {int i;
   // If flash already opened, return -1,
   if(at45db081_mbr.flags & fAT45DB081_OPENED){return(-1);}
   // Read page 0x0000 (sysinfo for check)
   at45db081_rd_page_buf1(0);
   // Copy cache to mbr-cache
   for(i=0;i<sizeof(at45db081_mbr);i++)
     {((char*)(&at45db081_mbr))[i]=at45db081_pcache.dpage[i];}
   // If flash is not formatted, return -1,
   // else return 0
   if(!strcmp_P(at45db081_mbr.marker,PSTR(FS_MARKER)))
     {// All OK!
      // Copy to cache page 1 (begin of file)
      at45db081_rd_page_buf1(1); 
      // File pointer to begin of file
      at45db081_mbr.fptr=0;
      // Flags
      at45db081_mbr.flags = fAT45DB081_OPENED;
      //
      return(0);
     }
   else
     {return(-1);}
  }
//----------------------------------------------------------------------
// Close flash. Synchornize cache and flash.
// Must be called before swich of power, if
// data was written to flash.
//	Return:
//		-1, if flash not open
//		0,  if all ok.
//----------------------------------------------------------------------
signed int at45db081_close()
  {int i;
   // Close device
   if(at45db081_mbr.flags & fAT45DB081_OPENED)
     {if(at45db081_mbr.flags & fAT45DB081_NEEDSYNC)
        {// Sychro current page
	 at45db081_wr_page_buf1(at45db081_mbr.curpage);
	 // Sychro current sysheader
	 // Copy to cache
         for(i=0;i<sizeof(at45db081_mbr);i++)
	   {at45db081_pcache.dpage[i]=((char*)(&at45db081_mbr))[i];}
	 // Page 0 sysinfo
	 at45db081_pcache.pagenum=0;
	 at45db081_pcache.crc_data=0;
	 at45db081_pcache.res0[0]='r';
	 at45db081_pcache.res0[1]='e';
	 at45db081_pcache.res0[2]='s';
	 at45db081_pcache.res0[3]='v';	 
	 // Write cache to flash
	 at45db081_wr_page_buf1(0);
	}
      // flags
      at45db081_mbr.flags &= ~(fAT45DB081_OPENED|fAT45DB081_NEEDSYNC);
      //
      return(0);
     }
   else
     {return(-1);
     }
  }
//----------------------------------------------------------------------
// Set file pointer to position pos.
// Return:
//	-1, if error;
//	current position, if all ok;
//----------------------------------------------------------------------
signed long int  at45db081_seek(signed long int pos)
  {if((at45db081_mbr.flags & fAT45DB081_OPENED)&&
      (pos<=at45db081_mbr.fsize))
     {if((pos >> 0x08)!=(at45db081_mbr.curpage-1))
        {// Sychro current page
         if(at45db081_mbr.flags & fAT45DB081_NEEDSYNC)
           {at45db081_wr_page_buf1(at45db081_mbr.curpage);
	    at45db081_mbr.flags &= ~fAT45DB081_NEEDSYNC;
	   }
	 // Read new current page
	 at45db081_rd_page_buf1((pos >> 0x08)+1);
	 //
	}
      // set current position
      at45db081_mbr.fptr = pos;
      //
      return(pos);
     }
   else
     {return(-1);
     }
  }
//----------------------------------------------------------------------
// Read from flash size bytes and put it to buf
// Return:
//	-1 - error (flash not open, for example)
//	number of read bytes, if all ok
//----------------------------------------------------------------------
signed long int  at45db081_read(char* buf, signed long int size)
  {signed long int i=0; // bytes counter
   signed long int fptr=at45db081_mbr.fptr; // file pointer
   signed long int fsize=at45db081_mbr.fsize; // file size
   // if flash opened
   if(at45db081_mbr.flags & fAT45DB081_OPENED)
     {while((i<size) && (fptr < fsize))
        {// Read from cache
	 buf[i]=at45db081_pcache.dpage[fptr & 0xFF];
	 // Change pointer
	 fptr++;
	 // increment bytes counter
	 i++;
	 // Check page end
	 if( (fptr & 0xFF)==0)
	   {// go to next page !!!
	    at45db081_seek(fptr);
	   }
	}
      // File pointer //at45db081_mbr.fptr=fptr;
      at45db081_seek(fptr);
      return(i);
     }
   else
     {return(-1);}
  }
//----------------------------------------------------------------------
// Write size bytes from buf to flash
// Return:
//	-1 - error (flash not open, for example)
//	number of written bytes, if all ok
//----------------------------------------------------------------------
signed long int  at45db081_write(char* buf, signed long int size)
  {signed long int i=0; // bytes counter
   signed long int fptr=at45db081_mbr.fptr; // file pointer
   // if flash opened
   if(at45db081_mbr.flags & fAT45DB081_OPENED)
     {while((i<size) && 
            (fptr < ((AT45DB081_PAGES-1)*AT45DB081_UPAGESIZE))
	   )
        {// Write to cache
	 at45db081_pcache.dpage[fptr & 0xFF]=buf[i];
	 // Change pointer
	 fptr++;
	 // increment bytes coynter
	 i++;
	 // File size
         if(fptr>at45db081_mbr.fsize){at45db081_mbr.fsize=fptr;}
	 // Check page end
	 if( (fptr & 0xFF)==0)
	   {// go to next page !!!
	    at45db081_mbr.flags |= fAT45DB081_NEEDSYNC;
	    at45db081_seek(fptr);
	   }
	}     
      // File pointer //at45db081_mbr.fptr=fptr;
      at45db081_seek(fptr);
      // Set synchro flag for close command
      at45db081_mbr.flags |= fAT45DB081_NEEDSYNC;
      return(i);
     }
   else
     {return(-1);
     }
  }
//----------------------------------------------------------------------
// Initalization chip and data structures
//----------------------------------------------------------------------
void at45db081_init()
  {int i;
   for(i=0;i<sizeof(at45db081_mbr);i++)
     {((char*)(&at45db081_mbr))[i]=0;}
  }
//----------------------------------------------------------------------
// Return file size, if flash opened
//----------------------------------------------------------------------
signed long int at45db081_fsize()
  {if(at45db081_mbr.flags & fAT45DB081_OPENED)
     {return(at45db081_mbr.fsize);
     }
   else
     {return(-1);}
  }
//----------------------------------------------------------------------
#endif
//----------------------------------------------------------------------
