//----------------------------------------------------------------------
//MODEULE: uart1.c
//COMMON DESCRIPTION:
//  This module provide support of internal UART0
//----------------------------------------------------------------------
#include "uart1.h"
#include "avr/signal.h"
#include "avr/pgmspace.h"
//----------------------------------------------------------------------
#ifdef UART1_ENABLED
  #ifdef UART1_BUFFERS_ENABLED
  //----------------------------------------------------------------------
  // RX Buffer
  unsigned char uart1_rxbuf[UART1_BUFFERS_LEN];
  unsigned char uart1_brxbuf; // begin of RX buffer
  unsigned char uart1_erxbuf; // end of RX buffer
  //----------------------------------------------------------------------
  // TX Buffer
  unsigned char uart1_txbuf[UART1_BUFFERS_LEN];
  unsigned char uart1_btxbuf; // begin of TX buffer
  unsigned char uart1_etxbuf; // end of TX buffer
  //----------------------------------------------------------------------
  // Functions of clearing buffers
  void uart1_clear_rx_buf(){uart1_brxbuf=uart1_erxbuf=0;}
  void uart1_clear_tx_buf(){uart1_btxbuf=uart1_etxbuf=0;}
  //----------------------------------------------------------------------
  // Returned TRUE if TX buffer is empty
  unsigned char uart1_txbuf_empty()
    {unsigned char ret=0;
     if(uart1_etxbuf==uart1_btxbuf){ret=0x01;}
     return(ret);
    }
  // Returned TRUE if RX buffer is empty
  unsigned char uart1_rxbuf_empty()
    {unsigned char ret=0;
     if(uart1_erxbuf==uart1_brxbuf){ret=0x01;}
     return(ret);
    }
  //----------------------------------------------------------------------
  // Function of cyclic increment
  unsigned char uart1_pcinc(unsigned char pc)
    {pc++;
     if(pc>=UART1_BUFFERS_LEN){pc=0;}
     return(pc);
    }
  //----------------------------------------------------------------------
  // Interrupt handlers
  //----------------------------------------------------------------------
  // RX Complete interrupt handler
  SIGNAL (SIG_UART1_RECV)
    {if(uart1_pcinc(uart1_erxbuf)!=uart1_brxbuf)
       {// Reading byte
        uart1_rxbuf[uart1_erxbuf]=input(UDR1);
	uart1_erxbuf=uart1_pcinc(uart1_erxbuf);
       }
    }
  //----------------------------------------------------------------------
  // TX Complete interrupt handler
  SIGNAL (SIG_UART1_TRANS)
    {if(uart1_etxbuf!=uart1_btxbuf)
       {// Writting byte
        output(UDR1,uart1_txbuf[uart1_btxbuf]);_NOP();_NOP();
        uart1_btxbuf=uart1_pcinc(uart1_btxbuf);
       }
    }
  #endif
  //----------------------------------------------------------------------
  // Init UART1
  void uart1_init()
    {// Init of UART1
     #ifndef UART1_SPD_CONST
       uart1_setspeed(UART1_DEF_SPD);
     #else
       // Mega-compats
       output(UBRR1H, (((long)F_CLK/(16*(long)UART1_SPD_CONST))-1) >> 0x08);_NOP();_NOP();
       output(UBRR1L, (((long)F_CLK/(16*(long)UART1_SPD_CONST))-1) &  0xFF);_NOP();_NOP();
     #endif
     //
     // Init buffers
     #ifdef UART1_BUFFERS_ENABLED
     uart1_clear_rx_buf();
     uart1_clear_tx_buf();
     // init UCR
     output(UCSR1B,(1<<RXEN) | (1<<TXEN) | (1<<RXCIE) | (1<<TXCIE));_NOP();_NOP();
     #else
     // init UCR
     output(UCSR1B,(1<<RXEN) | (1<<TXEN));_NOP();_NOP();
     #endif
     // Char size
     #ifdef UCSR1C
     output(UCSR1C, (3<<UCSZ00));_NOP();_NOP();
     #endif     
    }
  //----------------------------------------------------------------------
  // Define procedures
  #ifndef UART1_SPD_CONST
    // Set speed (speed in Bits/sec)
    void uart1_setspeed(unsigned int speed)
      {// Mega-compats
       unsigned int ubrr_i = (((long)F_CLK/(16*(long)speed))-1);
       output(UBRR1H, ubrr_i >> 0x08);_NOP();_NOP();
       output(UBRR1L, ubrr_i &  0xFF);_NOP();_NOP();
      }
  #endif
  //----------------------------------------------------------------------
  // Write to uart1 (or writing-buffer)
  unsigned char uart1_wr(char* c)
    {unsigned char fret;
     #ifdef UART1_BUFFERS_ENABLED
     unsigned char fstbyte=0;
     if(uart1_etxbuf==uart1_btxbuf){fstbyte=1;}
     // Write with buffer enable
     if(uart1_pcinc(uart1_etxbuf)!=uart1_btxbuf)
       {// Writting byte
        uart1_txbuf[uart1_etxbuf]=*c;
	uart1_etxbuf=uart1_pcinc(uart1_etxbuf);
	// Send first byte
	if((input(UCSR1A) & (1 << UDRE)) && fstbyte)
          {output(UDR1,*c);_NOP();_NOP();
	   uart1_btxbuf=uart1_pcinc(uart1_btxbuf);	  
          }
	 fret=1;
       }
     else
       {// Buffer is full, can't write
        fret=0;
       }
     #endif
     // Write without buffer enable
     #ifndef UART1_BUFFERS_ENABLED
     if(input(UCSR1A) & (1 << UDRE))
       {output(UDR1,*c);_NOP();_NOP();
        fret=1;
       }
     else{fret=0;}
     #endif
     return(fret);
    }
  //----------------------------------------------------------------------
  // Read from uart1 (or reading-buffer)
  unsigned char  uart1_rd(char* c)
    {unsigned char fret; 
     #ifdef UART1_BUFFERS_ENABLED
     // Read with buffer enable
     if(uart1_erxbuf!=uart1_brxbuf)
       {*c=uart1_rxbuf[uart1_brxbuf];
        uart1_brxbuf=uart1_pcinc(uart1_brxbuf);
	fret=1;
       }
     else
       {// Buffer is empty
        fret=0;
       }
     #endif
     // Read without buffer enable
     #ifndef UART1_BUFFERS_ENABLED
     if(input(UCSR1A) & (1 << RXC))
       {*c=input(UDR1);
        fret=1;
       }
     else{fret=0;}
     #endif
     return(fret);
    }
  //----------------------------------------------------------------------
  // Uart Strings Support
  //----------------------------------------------------------------------
  #ifdef UART1_STR_ENABLED
    // Write string from RAM
    void uart1_wrstr(unsigned char* s)
      {int i=0;
       while (s[i])
         {while(!uart1_wr(&s[i])){}
	  i++;
	 }
      }
    // Write string from Program Memory
    void uart1_p_wrstr(unsigned char* s)
      {int i=0;
       char ch;
       while (1)
         {ch=PRG_RDB(s+i);
	  if(!ch){break;}
	  while(!uart1_wr(&ch)){}
	  i++;
	 }
      }
    // Write number
    void uart1_wrint(unsigned int n)
      {unsigned char ch;
       unsigned int  d=10000;
       unsigned char k;
       //
       while (d)
         {k=n/d;
          ch=k+'0';
          n=n-d*k;
	  d=d/10;
	  while(!uart1_wr(&ch)){}
         }
       //
      }
    //
  #endif
#endif
